#ifndef __SAM9407_H
#define __SAM9407_H

/*
 *  Copyright (c) by Uros Bizjak <uros@kss-loka.si>
 *  Definitions for SAM9407 chips
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef __KERNEL__
#include <sound/driver.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <sound/core.h>

#include <sound/mpu401.h>
#include <sound/control.h>
#include <sound/hwdep.h>
#include <sound/info.h>
#include <sound/pcm.h>
#include <sound/rawmidi.h>

/* ------------------- DEFINES -------------------- */
#ifndef isis_t_magic
	#define isis_t_magic 0xa15a4f58
#endif

#ifndef sam9407_t_magic
	#define sam9407_t_magic 0xa15a4f59
#endif

#define SAM_ARRAYSIZE(x)	(sizeof(x)/sizeof(x[0]))

#define SAM9407P(sam, x)	((sam)->port + s_a_m_SAM9407##x)

#define s_a_m_SAM9407DATA8	0
#define s_a_m_SAM9407STATUS	1
#define s_a_m_SAM9407CONTROL	1
#define s_a_m_SAM9407DATA16	2

#define SAM_CTL_TABLE_SIZE	64   /* control table words */
#define SAM_CTL_TABLE_EXT_SIZE	12   /* control table extension words */

#define SAM_MAX_TRANSFER_SIZE	0x8000

/* commands */
#define SAM_CMD_EN_CONTROL	0xbe

/* acknowledges */
#define SAM_ACK			0xac
#define SAM_NAK			0xab

/* status byte bits */
#define SAM_STATUS_TE		0x80	/* transmit empty */
#define SAM_STATUS_RF		0x40	/* receiver full */

/* Memory mapping table IDs */
#define SAM_MMT_ID_MASK			0xF7F
#define  SAM_MMT_ID_MEMSIZE		0x00
#define  SAM_MMT_ID_FREE		0x01
#define  SAM_MMT_ID_RESERVED		0x02
#define  SAM_MMT_ID_MMT			0x03
#define  SAM_MMT_ID_MIDI_MT		0x04
#define  SAM_MMT_ID_MIDI_MT_EXT		0x05
#define  SAM_MMT_ID_CTRL_TBL		0x06
#define  SAM_MMT_ID_MIDI_PAR		0x10
#define  SAM_MMT_ID_MIDI_PCM		0x11
#define  SAM_MMT_ID_PCM			0x20
#define  SAM_MMT_ID_SYNTH		0x30
#define  SAM_MMT_ID_EOM			0xF7F
#define SAM_MMT_ROM_MASK		0x80

/* commands */
#define SAM_BOOT_CMD_UCODE_TRANSFER	0x0b /* undocummented boot command */
#define SAM_BOOT_CMD_UCODE_START	0x09 /* undocummented boot command */

#define SAM_CMD_UART_MOD	0x3f /* Switch to UART mode */
#define SAM_CMD_GEN_INT		0x48 /* Generate an interrupt */
#define SAM_CMD_HOT_RES		0x70 /* Hot reset */
#define   SAM_DATA_HOT_RES	0x11 /* Hot reset, hot reset */
#define   SAM_DATA_STOP_DSP	0x00 /* Hot reset, stop DSP */
#define   SAM_DATA_START_DSP	0x01 /* Hot reset, start DSP */

/* acknowledges */
#define SAM_ACK_CKSUM		0x10
#define SAM_ACK_IRQ		0x88
#define SAM_ACK_MOD		0xfe

/* PCM playback controls */
#define SAM_W_PITCH		0x44	/* Change the pitch of the wave during playback */
// different on ISIS
#define SAM_W_VOLLEFT	0x45	/* Change the current left volume */
#define SAM_W_VOLRIGHT	0x46	/* Change the current right volume */
#define SAM_W_VOLAUXLEFT	0x47 	/* Change the left auxiliary level */
#define SAM_W_VOLAUXRIGHT	0x49	/* Change the right auxiliary level */
#define SAM_W_FILT_FC	0x4a	/* Change filter cutoff frequency */
#define SAM_W_FILT_Q	0x4b	/* Change filter resonance */

#define SAM_END_XFER	0x43	/* Sent to SAM9407 after end of transfer */

#define SAM_MAX_PCM_CHANNELS	16
#define SAM_TARGET_12 	0
#define SAM_TARGET_34 	1
#define SAM_TARGET_56 	2
#define SAM_TARGET_78 	3


#define SAM_PLAY_OVERHEAD	16	/* magic playback buffer overhead */

#define SAM_MAX_BUFFER_SIZE	8192 //(32 * 1024)

#define SAM_HW_BUFF_SIZE_BYTES 8192 // actual buffer size

#define SAM_PCM_STEREO 	0x02
#define SAM_PCM_MONO	0x00
#define SAM_PCM_16BIT	0x00
#define SAM_PCM_8BIT	0x01

#define SAM_CMD_W_OPEN 	0x40
#define SAM_WAVE_ACK_OPEN 0x00
#define SAM_WAVE_ACK_CLOSE 0xC0

#define SAM_CMD_W_START 0x42
#define SAM_CMD_W_CLOSE 0x41
#define SAM_VOICE_PLAYING 1
#define SAM_VOICE_IN_USE 2
#define SAM_VOICE_START_BUSY 3
#define SAM_VOICE_STOP_BUSY 4

/* ------------------- STRUCTURES -------------------- */

typedef enum {
	SAM_ALIGN_NONE = 0,
	SAM_ALIGN_EVEN = 1,
} sam9407_align_t;

typedef enum {
	SAM_BOUNDARY_MASK_64K = (~((64 << 10) - 1)),
	SAM_BOUNDARY_MASK_256K = (~((256 << 10) - 1)),
} sam9407_boundary_t;

typedef struct {
	u16 id;
	u16 addr[2];
} sam9407_mmt_entry_t;

typedef struct _snd_sam9407 sam9407_t;
typedef struct _snd_sam9407_io_queue sam9407_io_queue_t;

typedef struct _snd_sam9407_voice sam9407_voice_t;
typedef struct _snd_sam9407_pcm sam9407_pcm_t;

typedef enum {
	SAM9407_NONE,
	SAM9407_PCM,
	SAM9407_SYNTH,
	SAM9407_MIDI
} sam9407_voice_type_t;

struct _snd_sam9407_io_queue {
	sam9407_t *sam;
	atomic_t status;
	spinlock_t lock;

	u8 ack;
	u8 *rcv_buffer;
	int rcv_bytes;
	int rcv_count;
	int expected_count;
	int error;
};

struct _snd_sam9407_voice {
	sam9407_t *sam;
	u8 number;
	u32 buf_addr;
	u32 status;
	u8 channel;
	snd_pcm_substream_t *substream;
	int running;

	u8 volume_left;
	u8 volume_right;
	u8 volume_aux_left;
	u8 volume_aux_right;
	u8 left_unmuted;
	u8 right_unmuted;
	u8 aux_left_unmuted;
	u8 aux_right_unmuted;


	snd_pcm_uframes_t hwptr;
	int use: 1,
	    pcm: 1,
	    synth: 1,
	    midi: 1;
};

struct _snd_sam9407_pcm {
	sam9407_t *sam;

};

struct _snd_sam9407 {

	/*
	Callbacks
	*/

	/* perform non-SAM reset actions */
	int (*reset) (sam9407_t *sam);

	/* low level SAM IO operations */

	void (*writeCtrl) (sam9407_t *sam, u8 ctrl);
	void (*writeData8) (sam9407_t *sam, u8 val);
	void (*writeData16) (sam9407_t *sam, u16 val);
	u8 (*readCtrl) (sam9407_t *sam);
	u8 (*readData8) (sam9407_t *sam);
	u16 (*readData16) (sam9407_t *sam);

	void (*writeData16Burst) (sam9407_t *sam, u16 *buffer, int count);
	// read the SAM in burst mode. buffer is assumed to be allocated
	void (*readData16Burst) (sam9407_t *sam, u16 *buffer, int count);

	/* interrupt control */
	int (*enableInterrupts) (sam9407_t *sam);
	int (*disableInterrupts) (sam9407_t *sam);

	/* Parent IO stuff */
//	u16 (*readParent)(sam9407_t *sam, u8 port);
//	u16 (*writeParent)(sam9407_t *sam, u8 port, u16 value);

	/* perform non-SAM init actions (e.g. clock setup) */
	int (*init) (sam9407_t *sam);
	int (*card_setup)(sam9407_t * sam, u32 * mem_size,
			  u16 * mmt_size);

	void *private_data;
	void (*private_free) (sam9407_t *sam);

	void (*allocatePCMBuffers)(sam9407_t*, snd_pcm_t *, u32 , u32);

	struct semaphore access_mutex;
	int hwdep_used;

	int ucode_loaded;

	sam9407_io_queue_t *midi_io_queue;	/* input queues */
	sam9407_io_queue_t *pcm_io_queue;
	sam9407_io_queue_t *synth_io_queue;
	sam9407_io_queue_t *system_io_queue;

	snd_card_t *card;
	snd_pcm_t *pcm;
	snd_pcm_substream_t *playback_substream;
	snd_pcm_substream_t *capture_substream;

	spinlock_t command_lock;
	spinlock_t transfer_lock;

	spinlock_t voice_lock;
	int voice_cnt;
	u16 voice_map;
	sam9407_voice_t voices[SAM_MAX_PCM_CHANNELS];

	/* memory mapping table shadow */
	spinlock_t mmt_lock;
	sam9407_mmt_entry_t *mmt;
	int mmt_entries;
	u32 mmt_addr;

	/* control table shadow */
	rwlock_t ctl_rwlock;
	u16 *ctl_table;

	struct list_head config_list;

	/* sound bank entries */
	int sbk_entries;

	wait_queue_head_t init_sleep;

	snd_info_entry_t *proc_entry;
	snd_info_entry_t *proc_entry_mmt;
	snd_info_entry_t *proc_entry_ctl;

	/* (Monitoring) volume table
	 * rows=input channel, cols = output channel,
	 * value & 0x7F = volume, value & 0x80 = mute
	 *
	 */
	u8 monitor_volume[8][4];
	u8 playback_volume[4];
	u8 record_volume[8];

	/* ISIS specific entries */
	u16 card_settings;

	/* MIDI interface */
	snd_rawmidi_t *rmidi_synth;
	snd_rawmidi_t *rmidi_midi;
	u8 midi_volume;
	u8 midi_pan;
	u8 midi_unmuted;

};

int snd_sam9407_create(snd_card_t * card, sam9407_t * _sam, sam9407_t ** rsam);
int snd_sam9407_free(sam9407_t *sam);
int snd_sam9407_dev_free(snd_device_t *device);

#if 0
/* initialization */
int snd_sam9407_create(snd_card_t * card,
		       unsigned long port,
		       int irq,
		       int (*mem_setup)(sam9407_t * chip, u32 * mem_size,
					u16 * mmt_size),
		       sam9407_t ** rchip);
#endif
int snd_sam9407_init(sam9407_t * chip);


/* I/O functions */
void snd_sam9407_interrupt(int irq, void *dev_id, struct pt_regs *regs);
int snd_sam9407_init_command(sam9407_t *sam, sam9407_io_queue_t * queue,
			     u8 *cmds, int cmds_bytes,
			     u8 *rcv_buffer, int rcv_bytes,
			     int ack);
int snd_sam9407_command(sam9407_t *sam, sam9407_io_queue_t * queue, u8 cmd,
			u8 *data, int data_len,
			u8 *rcv_buffer, int rcv_len,
			int ack);

int snd_sam9407_io_queue_create(sam9407_t *sam, sam9407_io_queue_t ** rqueue);

/* PCM async IRQ handler*/

void snd_sam9407_async_audio_irq(sam9407_t * sam, u8 data);

/* memory management */
int snd_sam9407_read(sam9407_t * sam, u16 *data,
		     u32 sam_addr, size_t wcount);
int snd_sam9407_write(sam9407_t * sam, u16 *data,
		      u32 sam_addr, size_t wcount);
int snd_sam9407_mmt_get_addr(sam9407_t * sam, u32 *addr);
int snd_sam9407_mmt_set_addr(sam9407_t * sam, u32 addr);

int snd_sam9407_mem_alloc(sam9407_t * sam, u16 id,
			  u32 * addr, size_t words,
			  sam9407_align_t align,
			  sam9407_boundary_t boundary_mask);
int snd_sam9407_mem_free(sam9407_t * sam, u16 id);
int snd_sam9407_mem_writestatictable(sam9407_t *sam);

/* control interface */
int snd_sam9407_ctl_get(sam9407_t * sam, u8 ctl, u8 *data);
int snd_sam9407_mixer(sam9407_t *sam);
int snd_sam9407_ctl_table_get(sam9407_t * sam, u8 ctl, u8 * data);

/* voice allocation */

/* MIDI uart */
int snd_sam9407_midi(sam9407_t *sam);

/* hwdep interface */
int snd_sam9407_ucode_hwdep_new(sam9407_t *sam, int device);

/* proc interface */
int snd_sam9407_proc_init(sam9407_t *sam);
int snd_sam9407_proc_done(sam9407_t *sam);

/* PCM interface */
int snd_sam9407_pcm(sam9407_t *sam, int device, int sam_target, snd_pcm_t ** rpcm);



#endif /* __KERNEL__ */

#define SAM_MAX_UCODE_LEN	0x10000		/* 64k words */

/* SAM9407 info */
typedef struct {
	int empty;
} sam9407_info_t;

typedef struct {
	char name[32];
	unsigned short len;			/* number of 16bit words */
	unsigned short csum;			/* checksum */
	unsigned short prog[SAM_MAX_UCODE_LEN];	/* 16bit words */
} sam9407_ucode_t;

typedef struct {
	char name[32];
	/* to be continued */
} sam9407_soundbank_t;

#define SNDRV_SAM9407_IOCTL_INFO	_IOR ('H', 0x40, sam9407_info_t)

#define SNDRV_SAM9407_IOCTL_FW_LOAD	_IOW ('H', 0x41, sam9407_ucode_t)

#define SNDRV_SAM9407_IOCTL_SBK_LOAD	_IOW ('H', 0x42, sam9407_soundbank_t)

#endif	/* __SAM9407_H */
