/*
 *  Copyright (c) by Uros Bizjak <uros@kss-loka.si>
 *  Lowlevel access to SAM9407 chip
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define __NO_VERSION__
#include <sound/driver.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <sound/core.h>
#include <sam9407.h>

#define OUTPUT_BUSY_LOOPS	100000


#define snd_sam9407_output_ready(sam)	(!(sam->readCtrl(sam) & SAM_STATUS_RF))

/* status byte ID values */
#define SAM_STATUS_ID_MASK	0x30
#define   SAM_STATUS_ID_MIDI	0x00
#define   SAM_STATUS_ID_PCM	0x10
#define   SAM_STATUS_ID_SYNTH	0x20
#define   SAM_STATUS_ID_SYSTEM	0x30

/* queue flags */
#define SAM_QUEUE_BUSY		0
#define SAM_QUEUE_WAIT_ACK	1
#define SAM_QUEUE_WAIT_MSG	2
#define SAM_QUEUE_WAIT_COUNT	3

/* max voices */
#define SAM_MAX_VOICES	33	/* 32 playback and 1 record channel */


/* ---------- */
//static inline int snd_sam9407_output_ready(sam9407_t * sam)	{
//	snd_printk("sam9407: snd_sam9407_output_ready?\n");
//	return (!(sam->readCtrl(sam) & SAM_STATUS_RF));
//}
/* send command byte to SAM9407 command port */
static inline int snd_sam9407_output_cmd(sam9407_t * sam, u8 cmd)
{
	int timeout;

	//snd_printk("sam9407: snd_sam9407_output_cmd = 0x%x\n", cmd);

	for (timeout = OUTPUT_BUSY_LOOPS; timeout; timeout--) {
		if (snd_sam9407_output_ready(sam)) {
			sam->writeCtrl(sam, cmd);
			return 0;
		}
	}
	snd_printk("sam9407: output cmd timeout, cmd = 0x%x\n", cmd);
	return -EIO;
}

/* send command stream to SAM9407 data port (in microcode initialization) */
static inline int snd_sam9407_output_cmds(sam9407_t * sam, u8 *cmds, int count)
{
	int err;
	//snd_printk("entering snd_sam9407_output_cmds (%d)\n",count);

	while (count-- > 0) {
		if ((err = snd_sam9407_output_cmd(sam, *cmds++)) < 0)
			return err;
	}
	return 0;
}

/* send data byte to SAM9407 data port */
static inline int snd_sam9407_output_byte(sam9407_t * sam, u8 data)
{
	//snd_printk("entering snd_sam9407_output_byte %02X\n",data);
	int timeout;

	for (timeout = OUTPUT_BUSY_LOOPS; timeout; timeout--) {
		if (snd_sam9407_output_ready(sam)) {
			sam->writeData8(sam, data);
			return 0;
		}
	}
	snd_printk("sam9407: output byte timeout, byte = 0x%x\n", data);
	return -EIO;
}

/* send data byte stream to SAM9407 data port */
static inline int snd_sam9407_output_bytes(sam9407_t * sam, u8 *bytes, int count)
{
	int err;

	while (count-- > 0) {
		if ((err = snd_sam9407_output_byte(sam, *bytes++)) < 0)
			return err;
	}
	return 0;
}

/* ---------- */

/* async audio IRQ handler */
#if 0
// currently implemented in the PCM routine
static int snd_sam9407_async_audio_irq(sam9407_t * sam, int data)
{
	/* TODO */
	snd_printk("sam9407: async audio IRQ received, channel %i\n", data);
	return 0;
}
#endif
/* main interrupt entry point */
void snd_sam9407_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	sam9407_t *sam = snd_magic_cast(sam9407_t, dev_id, return);
	sam9407_io_queue_t *queue;

	u8 stat, data;
	//snd_printk("entering sam interrupt\n");

	stat = sam->readCtrl(sam);
//	stat = inb(SAM9407P(sam, STATUS));

	/* check Transmit Empty status bit */
	if (stat & SAM_STATUS_TE) {
		return;
	}
	data = sam->readData8(sam);
//	data = inb(SAM9407P(sam, DATA8));

	//snd_printk("\t> sam9407: interrupt: stat = 0x%x, data = 0x%x\n", stat, data);

	switch (stat &= SAM_STATUS_ID_MASK) {
	case SAM_STATUS_ID_MIDI:
		queue = sam->midi_io_queue;
		break;
	case SAM_STATUS_ID_PCM:
		queue = sam->pcm_io_queue;
		break;
	case SAM_STATUS_ID_SYNTH:
		queue = sam->synth_io_queue;
		break;
	case SAM_STATUS_ID_SYSTEM:
		queue = sam->system_io_queue;
		break;
	default:
		/* not reached anyway */
		return;
	}

	if (test_bit(SAM_QUEUE_WAIT_ACK, &queue->status)) {
		if (data == queue->ack) {
			set_bit(SAM_QUEUE_WAIT_MSG, &queue->status);
			clear_bit(SAM_QUEUE_WAIT_ACK, &queue->status);
			goto __exit_ok;
		} else if (data == SAM_NAK) {
			/* read/write memory command can return NAK */
			queue->error = -EBUSY;
			clear_bit(SAM_QUEUE_BUSY, &queue->status);
			return;
		}
	}

	if (test_bit(SAM_QUEUE_WAIT_MSG, &queue->status)) {
		/* receive message */
		queue->rcv_buffer[queue->rcv_count++] = data;
	} else {
		/* process async IRQ (it can't arrive in WAIT_MSG state) */
		switch (stat)
		{
		case SAM_STATUS_ID_PCM:
			snd_sam9407_async_audio_irq(sam, data);
			return;
		case SAM_STATUS_ID_MIDI: /* use the generic MPU401 handler */
			if (sam->rmidi_midi) {
				snd_mpu401_uart_interrupt(irq,sam->rmidi_midi->private_data,regs);
			}
			return;
		}
		snd_printk("\t> sam9407: unexpected IRQ, id = 0x%x, data = 0x%x\n", stat, data);
		return;
	}
 __exit_ok:
	/* exit point for sync IRQ */
	if (queue->rcv_count == queue->rcv_bytes) {
		queue->error = 0;
                clear_bit(SAM_QUEUE_WAIT_MSG, &queue->status);
                clear_bit(SAM_QUEUE_BUSY, &queue->status);
	}
	return;
}


/* prepare queue to receive data */
static inline void snd_sam9407_queue_prepare(sam9407_io_queue_t * queue,
					     int rcv_bytes, u8 *rcv_buffer,
					     int ack)
{
	//snd_printk("entering queue_prepare\n");
	queue->rcv_count = 0;

	if (rcv_bytes > 0) {
		queue->rcv_bytes = rcv_bytes;
		queue->rcv_buffer = rcv_buffer;
	} else {
		queue->rcv_bytes = 0;
	}
	clear_bit(SAM_QUEUE_WAIT_MSG, &queue->status);

	if (ack >= 0) {
		queue->error = -EIO;
		queue->ack = ack & 0xff;
		set_bit(SAM_QUEUE_WAIT_ACK, &queue->status);
		set_bit(SAM_QUEUE_BUSY, &queue->status);
	} else {
		queue->error = 0;
		queue->ack = 0;
		clear_bit(SAM_QUEUE_WAIT_ACK, &queue->status);
	    set_bit(SAM_QUEUE_WAIT_MSG, &queue->status);
		set_bit(SAM_QUEUE_BUSY, &queue->status);
	}
	//snd_printk("queue prepared: status=%X ack=%X\n", queue->status, queue->ack);

}

/* transmit init command sequence to sam9407 */
int snd_sam9407_init_command(sam9407_t *sam, sam9407_io_queue_t * queue,
		         u8 *cmd_buffer, int cmd_bytes,
		         u8 *rcv_buffer, int rcv_bytes,
		         int ack)
{
	//snd_printk("entering init_command \n");

	//sam9407_t *sam = snd_magic_cast(sam9407_t, queue->sam, return -EINVAL);

	signed long end_time;
	int err;


	unsigned long flags;

	if (queue) {
		spin_lock(&queue->lock);
		//snd_printk("prepare queue\n");

		snd_sam9407_queue_prepare(queue, rcv_bytes, rcv_buffer, ack);
	}
	//snd_printk("output queue\n");
	spin_lock_irqsave(&sam->command_lock, flags);
	err = snd_sam9407_output_cmds(sam, cmd_buffer, cmd_bytes);
	spin_unlock_irqrestore(&sam->command_lock, flags);
	if (err)
		goto __exit_err;

	if (queue) {
		/* wait max 500ms in init mode (code initialization, etc..) */
		end_time = jiffies + (HZ / 2);
		do {
			if(test_bit(SAM_QUEUE_BUSY, &queue->status)) {
				set_current_state(TASK_UNINTERRUPTIBLE);
				schedule_timeout(1);
			} else {
				goto __exit_ok;
			}
		} while (end_time - (signed long) jiffies >= 0);
		snd_printk("sam9407: init command timeout\n");
 __exit_ok:
		err = queue->error;
	}
 __exit_err:
	/* cleanup mess */
	if (queue) {
		if (err) {
			clear_bit(SAM_QUEUE_WAIT_ACK, &queue->status);
			clear_bit(SAM_QUEUE_WAIT_MSG, &queue->status);
			clear_bit(SAM_QUEUE_BUSY, &queue->status);
		}
		spin_unlock(&queue->lock);
	}
	return err;
}

/* transmit command to sam9407 */
int snd_sam9407_command(sam9407_t *sam, sam9407_io_queue_t * queue, u8 cmd,
			u8 *data_buffer, int data_bytes,
			u8 *rcv_buffer, int rcv_bytes,
			int ack)
{
	//snd_printk("sam9407: snd_sam9407_command = 0x%x\n", cmd);

	signed long end_time;
	int loops;
	int err;

	unsigned long flags;

	if (queue) {
		spin_lock(&queue->lock);
		snd_sam9407_queue_prepare(queue, rcv_bytes, rcv_buffer, ack);
	}
	spin_lock_irqsave(&sam->command_lock, flags);
	err = snd_sam9407_output_cmd(sam, cmd);
	if (err)
		goto __io_error;
	err = snd_sam9407_output_bytes(sam, data_buffer, data_bytes);
 __io_error:
	spin_unlock_irqrestore(&sam->command_lock, flags);
	if (err)
		goto __exit_err;

	if (queue) {
		/* wait 2ms in fast path to receive data from sam9407 */
		for (loops = 200; loops; loops--) {
			if(test_bit(SAM_QUEUE_BUSY, &queue->status))
				udelay(10);
			else {
				// snd_printk("cmd loops = %i\n", loops);
				goto __exit_ok;
			}
		}
		/* wait for another 500ms - some commands need this */
		end_time = jiffies + (HZ);
		do {
			if(test_bit(SAM_QUEUE_BUSY, &queue->status)) {
				// FIXME: is this correct?
				set_current_state(TASK_UNINTERRUPTIBLE);
				schedule_timeout(1);
			} else {
				goto __exit_ok;
			}
		} while (end_time - (signed long) jiffies >= 0);
		snd_printk("sam9407: command 0x%x ack timeout\n", cmd);
 __exit_ok:
		err = queue->error;          
	}
 __exit_err:
	/* cleanup mess */
	if (queue) {
		if (err) {
			clear_bit(SAM_QUEUE_WAIT_ACK, &queue->status);
			clear_bit(SAM_QUEUE_WAIT_MSG, &queue->status);
			clear_bit(SAM_QUEUE_BUSY, &queue->status);
		}
		spin_unlock(&queue->lock);
	}
	return err;
}
