/*
 *  Copyright (c) by Uros Bizjak <uros@kss-loka.si>
 *  Memory management routines for SAM9407 chip
 *
 *  Based on code from sam9407-1.0.0 package (alloc.c),
 *  Copyright (C) 1998, 1999, 2000, 2001 Gerd Rausch <gerd@meshed.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define __NO_VERSION__
#include <sound/driver.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <sound/core.h>
#include <sam9407.h>


#define SAM_WORDS_SHIFT 0x01
#define SAM_WORDS_AVAIL 0x02

/* commands */
#define SAM_WRT_MEM		0x01 /* Initialize PC to SAM9407 transfer */
#define SAM_RD_MEM		0x02 /* Initialize SAM9407 to PC transfer */
#define SAM_GET_MMT		0x03 /* Get Memory Mapping Table address */
#define SAM_SET_MMT		0x04 /* Set Memory Mapping Table address */

/* acknowledges */
#define SAM_ACK_MMT		0

/* prototypes */
static size_t snd_sam9407_transfer_seg(sam9407_t * sam, u8 cmd,
				       u16 *data,
				       u32 sam_addr, size_t wcount);
static inline int _snd_sam9407_transfer(sam9407_t * sam, u8 cmd,
					u16 *data,
					u32 sam_addr, size_t wcount);

static inline void _snd_sam9407_mem_defrag(sam9407_t * sam);
static inline int _snd_sam9407_mem_alloc(sam9407_t * sam, size_t words,
					 sam9407_align_t align,
					 sam9407_boundary_t boundary_mask);

/* ---------- */

/* transfer to/from sam9407 memory segment */
static size_t snd_sam9407_transfer_seg(sam9407_t * sam, u8 cmd,
				       u16 *data,
				       u32 sam_addr, size_t wcount)
{
	u8 bytes[6];
	size_t max_wcount;

	int loops = 5;
	int err;


	/* limit transfer to page boundary */
	max_wcount = (~sam_addr & 0xffff) + 1;
	if (max_wcount > SAM_MAX_TRANSFER_SIZE)
		max_wcount = SAM_MAX_TRANSFER_SIZE;

	if (wcount > max_wcount)
		wcount = max_wcount;

	bytes[0] = sam_addr;		/* offset inside 64k page */
	bytes[1] = sam_addr >> 8;
	bytes[2] = sam_addr >> 16;	/* 64k page number */
	bytes[3] = sam_addr >> 24;
	bytes[4] = wcount;
	bytes[5] = wcount >> 8;

 __again:
	err = snd_sam9407_command(sam, sam->system_io_queue, cmd,
				  bytes, 6, NULL, 0, SAM_ACK);
	/* try 5 times to acquire transfer lock */
	if (err == -EBUSY) {
		if (--loops) {
			udelay(100);
			goto __again;
		}
	}

	if (err < 0)
		return 0;

	switch (cmd) {
	case SAM_WRT_MEM:
		//outsw(SAM9407P(sam, DATA16), data, wcount);
		sam->writeData16Burst(sam,data,wcount);
		break;
	case SAM_RD_MEM:
		//insw(SAM9407P(sam, DATA16), data, wcount);
		sam->readData16Burst(sam,data,wcount);
		break;
	default:
		break;
	}
	return wcount;
}

/* transfer to/from sam9407 memory */
static inline int _snd_sam9407_transfer(sam9407_t * sam, u8 cmd,
					u16 *data,
					u32 sam_addr, size_t wcount)
{
	size_t words;
	size_t cnt;

	cnt = wcount;
	while (cnt > 0) {
		words = snd_sam9407_transfer_seg(sam, cmd, data, sam_addr, cnt);
		/* something went wrong */
		if (words == 0) {
			snd_printk("sam9407: transfer %u words %s address 0x%x failed\n",
		  		   wcount, (cmd == SAM_WRT_MEM) ? "to" : "from", sam_addr);
			return -EIO;
		}
		data += words;
		sam_addr += words;
		cnt -= words;
	}
	return wcount;
}

/* ---------- */

/* read from SAM9407 memory */
int snd_sam9407_read(sam9407_t * sam, u16 *data,
		     u32 sam_addr, size_t wcount)
{
	return _snd_sam9407_transfer(sam, SAM_RD_MEM, data, sam_addr, wcount);
}

/* write to SAM9407 memory */
int snd_sam9407_write(sam9407_t * sam, u16 *addr,
		      u32 sam_addr, size_t wcount)
{
	return _snd_sam9407_transfer(sam, SAM_WRT_MEM, addr, sam_addr, wcount);
}

/* get MMT table address */
int snd_sam9407_mmt_get_addr(sam9407_t * sam, u32 *addr)
{

	u8 bytes = 0x00;
	u8 buffer[4];
	int err;

	err = snd_sam9407_command(sam, sam->synth_io_queue, SAM_GET_MMT,
				  &bytes, 1, buffer, 4, -1);
	if (err < 0)
		return err;

	*addr = (buffer[3] << 24) | (buffer[2] << 16) | (buffer[1] << 8) | buffer[0];
	return 0;
}

/* set MMT table address */
int snd_sam9407_mmt_set_addr(sam9407_t * sam, u32 addr)
{
	u8 bytes[4];

	int err;

	bytes[3] = addr >> 24;
	bytes[2] = addr >> 16;
	bytes[1] = addr >> 8;
	bytes[0] = addr;

	err = snd_sam9407_command(sam, NULL, SAM_SET_MMT,
				  bytes, 4, NULL, 0, -1);
	if (err < 0)
		return err;
	return 0;
}

/* ---------- */
static inline void _snd_sam9407_mem_defrag(sam9407_t * sam)
{
	sam9407_mmt_entry_t *mmt = sam->mmt;

	int slot;
	int target_slot, source_slot;
	int entries;

	for (slot = 0; slot < sam->mmt_entries; slot++) {
		if ((mmt[slot].id & SAM_MMT_ID_MASK) == SAM_MMT_ID_EOM)
			break;
		else if ((mmt[slot].id & SAM_MMT_ID_MASK) == SAM_MMT_ID_FREE) {
			target_slot = slot;

			/* preserve first free slot */
			target_slot++;

			for (source_slot = target_slot; source_slot < sam->mmt_entries; source_slot++) {
				if ((mmt[source_slot].id & SAM_MMT_ID_MASK) != SAM_MMT_ID_FREE)
					break;
			}

			/* move table entries */
			if(source_slot > target_slot) {
				entries = sam->mmt_entries - source_slot + 1;
				memmove(mmt + target_slot, mmt + source_slot,
					entries * sizeof(sam9407_mmt_entry_t));
			}
		}
	}
}

static inline int _snd_sam9407_mem_alloc(sam9407_t * sam, size_t words,
					 sam9407_align_t align,
					 sam9407_boundary_t boundary_mask)
{
	sam9407_mmt_entry_t *mmt = sam->mmt;
	int slot;

	u32 avail;

	u32 curr_addr, next_addr;
	u32 start_addr, end_addr;

	/* Keeps track of what we found */
	struct {
		int slot;
		u32 shift;
		u32 avail;
		u32 curr_addr, next_addr;
	} best;

	int entries;
	int status;
	int source_slot;
	int target_slot;

	end_addr = words - 1;
//	snd_printk("Searching for SAM mem block of %d words\n",words);

	/* words will not fit in one page */
	if (end_addr & boundary_mask) {
		//snd_printk("Buffer exceeds page\n");

		return -EINVAL;
	}

	best.avail = (unsigned int) (-1);	/* XXX MAX_?INT really */
	best.slot = -1;

	for (slot = 0; slot < sam->mmt_entries; slot++) {
		//snd_printk("examining slot %d of %d\n", slot,sam->mmt_entries);
		if ((mmt[slot].id & SAM_MMT_ID_MASK) == SAM_MMT_ID_EOM) {
			//snd_printk("End of memory detected at slot: %d\n", slot);
			break;
		}
		else if ((mmt[slot].id & SAM_MMT_ID_MASK) == SAM_MMT_ID_FREE) {
			curr_addr = (mmt[slot].addr[1] << 16) | mmt[slot].addr[0];
			//snd_printk("examining slot %d is free, address: %08X\n", slot, curr_addr);

			start_addr = curr_addr;
			/* align address */
			if (align) {
				start_addr += align;
				start_addr &= ~align;
			}
			//snd_printk(" aligned address: %08X\n", start_addr);

			/* if the space would exceed the page boundary,
			   let the address start at next page */
			end_addr = start_addr + words - 1;
			//snd_printk(" end address: %08X\n", end_addr);
			//snd_printk(" boundary mask: %08X\n", boundary_mask);


			if ((end_addr & boundary_mask) != (start_addr & boundary_mask)) {
				//snd_printk("  block would cross page boundary\n");
				start_addr = end_addr & boundary_mask;
				end_addr = start_addr;
				end_addr += words - 1;
			}
			//snd_printk(" Boundary check, address: from %08X to %08X \n", start_addr, end_addr);

			/* check slot size */
			next_addr = (mmt[slot + 1].addr[1] << 16) | mmt[slot + 1].addr[0];
			if (end_addr >= next_addr) {
				//snd_printk(" not enough space in this slot\n");
				continue;
			}

			avail = next_addr - end_addr;
			if (avail < best.avail) {
				//snd_printk(" this one (%d) is the best available slot\n", slot);
				best.slot = slot;
				best.shift = start_addr - curr_addr;
				best.avail = avail;
				best.curr_addr = curr_addr;
				best.next_addr = next_addr;
			}
		}
	}

	/* card ran out of memory */
	if (best.slot < 0) {
		snd_printk("No more free space in MMT\n");
		return -ENOMEM;
	}

	/* end_addr points to last word address, adjust avail */
	best.avail--;

	/* number of entries to move */
	entries = slot - best.slot + 1;

	status = 0;
	source_slot = best.slot;
	target_slot = best.slot;

	/* mark non-aligned words free */
	if (best.shift != 0) {
		/* split entry, check space in MM table */
		slot++;
		if (slot == sam->mmt_entries)
			return -ENOSPC;

		/* split entry and shift best.slot */
		status |= SAM_WORDS_SHIFT;
		best.slot++;
		target_slot++;
	}

	/* mark available words free */
	if (best.avail != 0) {
		/* split entry, check space in MM table */
		slot++;
		if (slot == sam->mmt_entries)
			return -ENOSPC;

		/* split entry */
		status |= SAM_WORDS_AVAIL;
		target_slot++;
	}

	/* size fits exactly, return */
	if (!status)
		return best.slot;

	/* move table entries */
	memmove(mmt + target_slot, mmt + source_slot,
		entries * sizeof(sam9407_mmt_entry_t));

	/* write down shifted entry start address */
	if (status & SAM_WORDS_SHIFT) {
		curr_addr = best.curr_addr + best.shift;
		mmt[best.slot].addr[1] = curr_addr >> 16;
		mmt[best.slot].addr[0] = curr_addr;
	}

	/* write down splitted entry start address */
	if (status & SAM_WORDS_AVAIL) {
		curr_addr = best.next_addr - best.avail;
		mmt[best.slot + 1].addr[1] = curr_addr >> 16;
		mmt[best.slot + 1].addr[0] = curr_addr;
	}

	/* returned slot id is undefined */
	return best.slot;
}

int snd_sam9407_mem_alloc(sam9407_t * sam, u16 id,
			  u32 * addr, size_t words,
			  sam9407_align_t align,
			  sam9407_boundary_t boundary_mask)
{
	sam9407_mmt_entry_t *mmt = sam->mmt;
	int slot;
	int retry = 0;
	int err;

	snd_printk("Trying to allocate %d words of SAM memory, id= %04X\n",words,id);
 __again:
	slot = _snd_sam9407_mem_alloc(sam, words, align, boundary_mask);
	if (slot < 0) {
		if (retry) {
			return slot;
		} else {
			retry++;
			snd_printk("defrag needed\n");
			_snd_sam9407_mem_defrag(sam);

			goto __again;
		}
	} else {
		mmt[slot].id = id;
		err = snd_sam9407_write(sam, (u16 *) mmt,
					sam->mmt_addr, sam->mmt_entries * 3);
  		snd_printk("allocated to slot %d: id= %04X addr=%08X\n",slot,id,(mmt[slot].addr[1] << 16) | mmt[slot].addr[0] );

		if (addr)
			*addr = (mmt[slot].addr[1] << 16) | mmt[slot].addr[0];
		return err;
	}
}

int snd_sam9407_mem_free(sam9407_t * sam, u16 id)
{
	sam9407_mmt_entry_t *mmt = sam->mmt;
	int slot;
	int err;
	snd_printk("Freeing SAM memory id= %04X\n",id);

	if (id & SAM_MMT_ROM_MASK)
		return -EINVAL;
	if (id == SAM_MMT_ID_RESERVED)
		return -EINVAL;
	for (slot = 0; slot < sam->mmt_entries; slot++) {
		if ((mmt[slot].id & SAM_MMT_ID_MASK) == SAM_MMT_ID_EOM)
			break;
		else if ((mmt[slot].id & SAM_MMT_ID_MASK) == id) {
			mmt[slot].id = SAM_MMT_ID_FREE;
			err = snd_sam9407_write(sam, (u16 *) mmt,
						sam->mmt_addr, sam->mmt_entries * 3);

			return err;
		}
	}
	return -EINVAL;
}

/*
This function writes the static MMT used by the ISIS windows drivers
Currently, we'd better use this table before trying to allocate dynamically. (get it working first)

0x000, 0x200000
0x002, 0x0
0x006, 0x681e
0x002, 0x685e
0x003, 0x6b19
0x001, 0x6cc1
0x002, 0x6cc2

0x020, 0x12D00
0x120, 0x14D10
0x820, 0x16D20
0x920, 0x18D20
0xA20, 0x1AD20
0xB20, 0x1CD20
0x001, 0x1ED20

0x002, 0x200000
0x002, 0x1000000
0x082, 0x1800000
0x0ffff, 0x200000

*/
int snd_sam9407_mem_writestatictable(sam9407_t *sam) {

	u16 id;
	size_t words;
	int err;
	u32 buf_addr;

	int err1=0;

	snd_printk(" statictable: Preallocating MMT PCM entries\n");


	words=SAM_HW_BUFF_SIZE_BYTES + SAM_PLAY_OVERHEAD;
/*
	id = SAM_MMT_ID_PCM | (channel_number << 8);
*/
	id = SAM_MMT_ID_PCM | (0 << 8);
 if ((err = snd_sam9407_mem_alloc(sam, id, &buf_addr, words, SAM_ALIGN_EVEN, SAM_BOUNDARY_MASK_64K))<0) {
		snd_printk(" statictable: could not allocate %03X\n",id);err1=-1;
	} else {
		//snd_printk(" statictable: allocated %03X to %08X\n",id, buf_addr);
	}

	id = SAM_MMT_ID_PCM | (1 << 8);
 if ((err = snd_sam9407_mem_alloc(sam, id, &buf_addr, words, SAM_ALIGN_EVEN, SAM_BOUNDARY_MASK_64K))<0) {
		snd_printk(" statictable: could not allocate %03X\n",id);err1=-1;
	} else {
		//snd_printk(" statictable: allocated %03X to %08X\n",id, buf_addr);
	}
#if 0
	// record buffer don't seem to need the overhead memory
	words=SAM_HW_BUFF_SIZE_BYTES;


	id = SAM_MMT_ID_PCM | (8 << 8);
 if ((err = snd_sam9407_mem_alloc(sam, id, &buf_addr, words, SAM_ALIGN_EVEN, SAM_BOUNDARY_MASK_64K))<0) {
		snd_printk(" statictable: could not allocate %03X\n",id);err1=-1;
	} else {
		//snd_printk(" statictable: allocated %03X to %08X\n",id, buf_addr);
	}

	id = SAM_MMT_ID_PCM | (9 << 8);
 if ((err = snd_sam9407_mem_alloc(sam, id, &buf_addr, words, SAM_ALIGN_EVEN, SAM_BOUNDARY_MASK_64K))<0) {
		snd_printk(" statictable: could not allocate %03X\n",id);err1=-1;
	} else {
		//snd_printk(" statictable: allocated %03X to %08X\n",id, buf_addr);
	}

	id = SAM_MMT_ID_PCM | (10 << 8);
 if ((err = snd_sam9407_mem_alloc(sam, id, &buf_addr, words, SAM_ALIGN_EVEN, SAM_BOUNDARY_MASK_64K))<0) {
		snd_printk(" statictable: could not allocate %03X\n",id);err1=-1;
	} else {
		//snd_printk(" statictable: allocated %03X to %08X\n",id, buf_addr);
	}

	id = SAM_MMT_ID_PCM | (11 << 8);
 if ((err = snd_sam9407_mem_alloc(sam, id, &buf_addr, words, SAM_ALIGN_EVEN, SAM_BOUNDARY_MASK_64K))<0) {
		snd_printk(" statictable: could not allocate %03X\n",id);err1=-1;
	} else {
		//snd_printk(" statictable: allocated %03X to %08X\n",id, buf_addr);
	}
#endif

	return err1;
}
